import math

import cv2
import numpy as np

from zebrazoom.code.deepLearningFunctions.labellingFunctions import drawWhitePointsOnInitialImages, saveImagesAndData


class _FindTailExtremityMixin:
  @staticmethod
  def __initialiseDistance2(distance, boundary):
    TotalBPts   = len(boundary)
    distance[0] = 0
    for i in range(1, TotalBPts+1):
      if (i == TotalBPts):
        Pt = boundary[0][0]
      else:
        Pt = boundary[i][0]
      AvantPt = boundary[i-1][0]
      Dx = AvantPt[0] - Pt[0]
      Dy = AvantPt[1] - Pt[1]
      if i:
        distance[i] = distance[i-1] + math.sqrt(Dx*Dx + Dy*Dy)
      else:
        distance[i] = math.sqrt(Dx*Dx + Dy*Dy)
    return [distance[TotalBPts], distance]

  @staticmethod
  def __calculateJuge(indice, distance, max):
    juge = 0
    if distance[indice] < max - distance[indice]:
      juge = (max-2*distance[indice])/max
    else:
      juge = (2*distance[indice]-max)/max
    return juge

  @staticmethod
  def __calculateJuge2(indice, distance, bord1, bord2, nb):

    dist  = 0
    dist2 = 0
    dist3 = 0
    dist4 = 0
    mindist = 10000000000

    if indice < bord1:
      dist  = distance[bord1] - distance[indice]
      dist2 = distance[indice] + (distance[nb] - distance[bord1])
    else:
      dist  = distance[indice] - distance[bord1]
      dist2 = distance[bord1] + ( distance[nb] - distance[indice] )

    if indice < bord2:
      dist3 = distance[bord2] - distance[indice]
      dist4 = distance[indice] + ( distance[nb] - distance[bord2] )
    else:
      dist3 = distance[indice] - distance[bord2]
      dist4 = distance[bord2] + ( distance[nb] - distance[indice] )

    if dist < mindist:
      mindist = dist

    if dist2 < mindist:
      mindist = dist2

    if dist3 < mindist:
      mindist = dist3

    if dist4 < mindist:
      mindist = dist4

    return mindist

  def __insideTailExtremete(self, distance, DotProds, max_droite, min_gauche, max_bas, min_haut, ind_droite, ind_gauche, ind_bas, ind_haut, tailRange, boundary, dst):
    TotalBPts = len(boundary)

    # This may require some adjustements in the future (maybe some value other than 25)
    dist_calculate_curv = int(TotalBPts / 25)
    if dist_calculate_curv < 3:
      dist_calculate_curv = 3

    max = 0

    AheadPtr  = 0
    BehindPtr = 0
    Ptr       = 0

    juge = 0
    x = 0
    y = 0

    for i in tailRange:
      AheadPtr = (i + dist_calculate_curv) % TotalBPts
      BehindPtr = (i + TotalBPts - dist_calculate_curv) % TotalBPts
      Ptr = i % TotalBPts
      AheadPt = boundary[AheadPtr][0]
      Pt = boundary[Ptr][0]

      BehindPt    = boundary[BehindPtr][0]
      AheadVec    = [AheadPt[0] - Pt[0],  AheadPt[1] - Pt[1]]

      BehindVec   = [Pt[0] - BehindPt[0], Pt[1] - BehindPt[1]]
      DotProdVal  = (AheadVec[0])*(BehindVec[0]) + (AheadVec[1])*(BehindVec[1])
      DotProds[i] = DotProdVal
      x = Pt[0]
      y = Pt[1]

      # Hmm... not sure about this part bellow...
      fin_boucle = tailRange[len(tailRange)-1]
      juge = self.__calculateJuge(i,distance,distance[fin_boucle-1])
      # The line above should probably be replace by the one below at some point !!!
      # juge=self.__calculateJuge2(i,distance,bord1,bord2,TotalBPts);
      # (juge > trackParameters.minDistFromTailExtremityToTailBasis)

      if x > max_droite:
        max_droite = x
        ind_droite = i
      if x < min_gauche:
        min_gauche = x
        ind_gauche = i
      if y > max_bas:
        max_bas = y
        ind_bas = i
      if (y < min_haut): # and (juge < 0.20):
        min_haut = y
        ind_haut = i

      droite = ind_droite
      gauche = ind_gauche
      haut   = ind_haut
      bas    = ind_bas

      max = distance[i]

    return [max, max_droite, min_gauche, max_bas, min_haut, ind_droite, ind_gauche, ind_bas, ind_haut]

  def _findTailExtremete(self, rotatedContour, bodyContour, aaa, bord1b, bord2b, debug, dst, tailExtremityMaxJugeDecreaseCoeff):
    max  = 0
    max2 = 0
    TotalBPts = len(rotatedContour)
    DotProds = np.zeros(TotalBPts)
    distance = np.zeros(TotalBPts)

    for i in range(0, TotalBPts):
      distance[i] = 0

    distance2 = np.zeros(TotalBPts+1)
    for i in range(0, TotalBPts):
      distance2[i] = 0

    [d, distance2] = self.__initialiseDistance2(distance2, rotatedContour)

    max_droite = 0
    min_gauche = 5000
    max_bas    = 0
    min_haut   = 5000

    ind_droite = 0
    ind_gauche = 0
    ind_bas    = 0
    ind_haut   = 0

    x = 0
    y = 0

    bord1 = 0
    bord2 = 0
    if (bord2b < bord1b):
      bord1 = bord2b
      bord2 = bord1b
    else:
      bord1 = bord1b
      bord2 = bord2b

    Bord1 = rotatedContour[bord1][0]
    Bord2 = rotatedContour[bord2][0]

    max1 = distance2[bord2] - distance2[bord1]
    max2 = (distance2[bord1] - distance2[0])  + (distance2[len(rotatedContour)] - distance2[bord2])

    if self._hyperparameters["checkAllContourForTailExtremityDetect"] == 0:
      tailRange = []
      if (max1 > max2):
        for i in range(bord1, bord2):
          tailRange.append(i)
      else:
        for i in range(0, bord1):
          tailRange.append(i)
        for i in range(bord2, len(rotatedContour)):
          tailRange.append(i)
    else:
      tailRange = []
      for i in range(0, len(rotatedContour)):
        tailRange.append(i)

    [max2, max_droite, min_gauche, max_bas, min_haut, ind_droite, ind_gauche, ind_bas, ind_haut] = self.__insideTailExtremete(distance2, DotProds, max_droite, min_gauche, max_bas, min_haut, ind_droite, ind_gauche, ind_bas, ind_haut, tailRange, rotatedContour, dst)

    MostCurvy = 100000
    CurrentCurviness = 0
    MostCurvyIndex = 0
    TailIndex = 0

    max_dist = 15000.0

    jugeDroite = self.__calculateJuge2(ind_droite,distance2,bord1,bord2,TotalBPts)
    jugeGauche = self.__calculateJuge2(ind_gauche,distance2,bord1,bord2,TotalBPts)
    jugeHaut   = self.__calculateJuge2(ind_haut,distance2,bord1,bord2,TotalBPts)
    jugeBas    = self.__calculateJuge2(ind_bas,distance2,bord1,bord2,TotalBPts)
    maxJuge    = 0.0
    if jugeDroite > jugeGauche:
      maxJuge = jugeDroite
    else:
      maxJuge = jugeGauche

    if jugeHaut > maxJuge:
      maxJuge = jugeHaut

    if jugeBas > maxJuge:
      maxJuge = jugeBas

    maxJuge = maxJuge - tailExtremityMaxJugeDecreaseCoeff * maxJuge

    if debug:
      print("MostCurvy:",MostCurvy,";maxJuge:",maxJuge)

    DotProdPtr = DotProds[ind_droite]
    if debug:
      print("Droite (red) = curv: ", DotProdPtr, " ; jugeDroite: ", jugeDroite)

    if ((DotProdPtr < MostCurvy) and (jugeDroite > maxJuge)):
      MostCurvy =  DotProdPtr
      MostCurvyIndex = ind_droite
      if debug:
        print("droite wins")

    DotProdPtr=DotProds[ind_gauche]
    if (debug):
      print("Gauche (blue) = curv: ", DotProdPtr, " ; jugeGauche: ", jugeGauche)

    if (( DotProdPtr < MostCurvy) and (jugeGauche > maxJuge)):
      MostCurvy =  DotProdPtr
      MostCurvyIndex = ind_gauche
      if (debug):
        print("gauche wins")

    DotProdPtr = DotProds[ind_haut]
    if debug:
      print("Haut (white) = curv: ", DotProdPtr, " ; jugeHaut: ", jugeHaut)

    if (( DotProdPtr < MostCurvy) and (jugeHaut > maxJuge) and self._hyperparameters["considerHighPointForTailExtremityDetect"]):
      MostCurvy =  DotProdPtr
      MostCurvyIndex = ind_haut
      if (debug):
        print("haut wins")

    DotProdPtr = DotProds[ind_bas]
    if debug:
      print("Bas (Purple)= curv: ", DotProdPtr, " ; jugeBas: ", jugeBas)

    if (( DotProdPtr < MostCurvy) and (jugeBas > maxJuge)):
      MostCurvy =  DotProdPtr
      MostCurvyIndex = ind_bas
      if (debug):
        print("bas wins")

    if debug:
      # Droite
      pt1 = bodyContour[int(ind_droite)][0]
      cv2.circle(dst, (pt1[0],pt1[1]), 1, (255, 0, 0), -1)
      # Gauche
      pt1 = bodyContour[int(ind_gauche)][0]
      cv2.circle(dst, (pt1[0],pt1[1]), 1, (0, 255, 0), -1)
      # Haut
      pt1 = bodyContour[int(ind_haut)][0]
      cv2.circle(dst, (pt1[0],pt1[1]), 1, (0, 0, 255), -1)
      # Bas
      pt1 = bodyContour[int(ind_bas)][0]
      cv2.circle(dst, (pt1[0],pt1[1]), 1, (255, 255, 0), -1)
      if False: # The following can sometimes be useful when debugging
        for i in range(0, len(rotatedContour)):
          pt1 = rotatedContour[int(i)][0]
          cv2.circle(dst, (pt1[0],pt1[1]), 1, (0, 0, 0), -1)
      if self._hyperparameters["debugTrackingPtExtremeLargeVerticals"]:
        dst = dst[pt1[1]-200:len(dst), :]
      # Plotting points
      self._debugFrame(dst, title='Frame')

    # allPossibilities = [[ind_droite,DotProds[ind_droite],jugeDroite], [ind_gauche,DotProds[ind_gauche],jugeGauche], [ind_haut,DotProds[ind_haut],jugeHaut], [ind_bas,DotProds[ind_bas],jugeBas]]

    return [MostCurvyIndex, distance2]


class TailTrackingExtremityDetectMixin(_FindTailExtremityMixin):
  def __checkIfMidlineIsInBlob(self, headX, headY, pointAlongTheTail, bodyContour, dst, size, distance2, debugAdv):
    tail = self._getMidline(headX, headY, pointAlongTheTail, bodyContour, dst, size, distance2, debugAdv)

    tail2 = tail[0]
    n = len(tail2)
    allMidlinePointsInsideBlob = True
    for j in range(0, n):
      dist = cv2.pointPolygonTest(bodyContour,(tail2[j][0],tail2[j][1]),False)
      if dist < 0:
        allMidlinePointsInsideBlob = False

    tailLength = 0
    if allMidlinePointsInsideBlob:
      for j in range(0, n-1):
        tailLength = tailLength + math.sqrt( pow(tail2[j,0]-tail2[j+1,0], 2) + pow(tail2[j,1]-tail2[j+1,1], 2) )

    return [allMidlinePointsInsideBlob, tailLength]

  def __findTheTwoSides(self, headPosition, bodyContour, dst):
    if self._hyperparameters["detectMouthInsteadOfHeadTwoSides"] == 1:

      lenX = len(dst[0])
      lenY = len(dst)
      originalShape = np.zeros((lenY, lenX))
      originalShape[:, :] = 0
      originalShape = originalShape.astype(np.uint8)
      cv2.fillPoly(originalShape, pts =[bodyContour], color=(255))

      # Heading calculation: first approximation
      minWhitePixel = 1000000000
      bestAngle     = 0
      nTries        = 50
      for i in range(0, nTries):
        angleOption = i * ((2 * math.pi) / nTries)
        startPoint = (int(headPosition[0]), int(headPosition[1]))
        endPoint   = (int(headPosition[0] + 100000 * math.cos(angleOption)), int(headPosition[1] + 100000 * math.sin(angleOption)))
        testImage  = originalShape.copy()
        testImage  = cv2.line(testImage, startPoint, endPoint, (0), 1)
        nbWhitePixels = cv2.countNonZero(testImage)
        if nbWhitePixels < minWhitePixel:
          minWhitePixel = nbWhitePixels
          bestAngle     = angleOption
      bestAngleAfterFirstStep = bestAngle

      # Heading calculation: second (and refined) approximation
      # Searching for the optimal value of iterationsForErodeImageForHeadingCalculation
      countTries = 0
      nbIterations2nbWhitePixels = {}
      if "iterationsForErodeImageForHeadingCalculation" in self._hyperparameters:
        iterationsForErodeImageForHeadingCalculation = self._hyperparameters["iterationsForErodeImageForHeadingCalculation"]
      else:
        iterationsForErodeImageForHeadingCalculation = 4
      kernel = np.ones((3, 3), np.uint8)
      nbWhitePixelsMax = 75
      while (iterationsForErodeImageForHeadingCalculation > 0) and (countTries < 50) and not(iterationsForErodeImageForHeadingCalculation in nbIterations2nbWhitePixels):
        testImage2 = cv2.erode(testImage, kernel, iterations = iterationsForErodeImageForHeadingCalculation)
        nbWhitePixels = cv2.countNonZero(testImage2)
        nbIterations2nbWhitePixels[iterationsForErodeImageForHeadingCalculation] = nbWhitePixels
        if nbWhitePixels < nbWhitePixelsMax:
          iterationsForErodeImageForHeadingCalculation = iterationsForErodeImageForHeadingCalculation - 1
        if nbWhitePixels >= nbWhitePixelsMax:
          iterationsForErodeImageForHeadingCalculation = iterationsForErodeImageForHeadingCalculation + 1
        countTries = countTries + 1
      best_iterations = 0
      minDist = 10000000000000
      for iterations in nbIterations2nbWhitePixels:
        nbWhitePixels = nbIterations2nbWhitePixels[iterations]
        dist = abs(nbWhitePixels - nbWhitePixelsMax)
        if dist < minDist:
          minDist = dist
          best_iterations = iterations
      iterationsForErodeImageForHeadingCalculation = best_iterations
      self._hyperparameters["iterationsForErodeImageForHeadingCalculation"] = iterationsForErodeImageForHeadingCalculation

      maxDist = -1
      for i in range(0, nTries):
        angleOption = bestAngleAfterFirstStep - (math.pi / 5) + i * ((2 * (math.pi / 5)) / nTries)

        startPoint = (int(headPosition[0]), int(headPosition[1]))
        endPoint   = (int(headPosition[0] + 100000 * math.cos(angleOption)), int(headPosition[1] + 100000 * math.sin(angleOption)))
        testImage  = originalShape.copy()

        # applying dilation with the iterationsForErodeImageForHeadingCalculation value found
        testImage = cv2.erode(testImage, kernel, iterations = iterationsForErodeImageForHeadingCalculation)
        testImage  = cv2.line(testImage, startPoint, endPoint, (0), 1)
        nbWhitePixels = cv2.countNonZero(testImage)
        if nbWhitePixels < minWhitePixel:
          minWhitePixel = nbWhitePixels
          bestAngle     = angleOption

      # Finding the 'mouth' of the fish
      unitVector = np.array([math.cos(bestAngle + math.pi), math.sin(bestAngle + math.pi)])
      factor     = 1
      headPos    = np.array(headPosition)
      testBorder = headPos + factor * unitVector
      testBorder = testBorder.astype(int)
      while (cv2.pointPolygonTest(bodyContour, (testBorder[0], testBorder[1]), True) > 0) and (factor < 100) and (testBorder[0] >= 0) and (testBorder[1] >= 0) and (testBorder[0] < len(dst[0])) and (testBorder[1] < len(dst)):
        factor = factor + 1
        testBorder = headPos + factor * unitVector

      # Finding the indexes of the two "border points" along the contour (these are the two points that are the closest from the 'mouth' of fish)
      xOtherBorder = testBorder[0]
      yOtherBorder = testBorder[1]
      minDist1 = 1000000000000
      minDist2 = 1000000000000
      indMin1  = 0
      indMin2  = 0
      for i in range(0, len(bodyContour)):
        Pt   = bodyContour[i][0]
        dist = math.sqrt((Pt[0] - xOtherBorder)**2 + (Pt[1] - yOtherBorder)**2)
        if (dist < minDist1):
          minDist2 = minDist1
          indMin2  = indMin1
          minDist1 = dist
          indMin1  = i
        else:
          if (dist < minDist2):
            minDist2 = dist
            indMin2  = i

      res = [indMin1, indMin2, bestAngle + math.pi]

    else:

      res = np.zeros(2)

      x = headPosition[0]
      y = headPosition[1]

      minDist = 1000000000000
      indMin  = 0
      for i in range(0, len(bodyContour)):
        Pt   = bodyContour[i][0]
        dist = math.sqrt((Pt[0] - x)**2 + (Pt[1] - y)**2)
        if (dist < minDist):
          minDist = dist
          indMin  = i

      res[0] = indMin
      PtClosest = bodyContour[indMin][0]
      headPos   = np.array(headPosition)

      unitVector = np.array([x - PtClosest[0], y - PtClosest[1]])
      unitVectorLength = math.sqrt(unitVector[0]**2 + unitVector[1]**2)
      unitVector[0] = unitVector[0] / unitVectorLength
      unitVector[1] = unitVector[1] / unitVectorLength

      factor = 1
      testBorder = headPos + factor * unitVector
      testBorder = testBorder.astype(int)
      while (cv2.pointPolygonTest(bodyContour, (testBorder[0], testBorder[1]), True) > 0) and (factor < 100) and (testBorder[0] >= 0) and (testBorder[1] >= 0) and (testBorder[0] < len(dst[0])) and (testBorder[1] < len(dst)):
        factor = factor + 1
        testBorder = headPos + factor * unitVector

      xOtherBorder = testBorder[0]
      yOtherBorder = testBorder[1]

      minDist = 1000000000000
      indMin2  = 0
      for i in range(0, len(bodyContour)):
        Pt   = bodyContour[i][0]
        dist = math.sqrt((Pt[0] - xOtherBorder)**2 + (Pt[1] - yOtherBorder)**2)
        if (dist < minDist):
          minDist = dist
          indMin2  = i

      res[1] = indMin2

    if False:
      cv2.circle(dst, (pt1[0],pt1[1]), 1, (0, 0, 255), -1)
      cv2.circle(dst, (pt2[0],pt2[1]), 1, (0, 0, 255), -1)
      self._debugFrame(dst, title='Frame')

    return res

  def __findBodyContour(self, headPosition, thresh1, initialCurFrame, back, wellNumber=-1, frameNumber=-1):
    if self._hyperparameters["saveBodyMask"] and self._hyperparameters["bodyMask_addWhitePoints"]:
      [img, thresh1] = drawWhitePointsOnInitialImages(initialCurFrame, back, self._hyperparameters)

    thresh1[:,0] = 255
    thresh1[0,:] = 255
    thresh1[:, len(thresh1[0])-1] = 255
    thresh1[len(thresh1)-1, :]    = 255

    x = headPosition[0]
    y = headPosition[1]
    cx = 0
    cy = 0
    takeTheHeadClosestToTheCenter = 1
    bodyContour = 0

    if self._hyperparameters["findContourPrecision"] == "CHAIN_APPROX_SIMPLE":
      contourPrecision = cv2.CHAIN_APPROX_SIMPLE
    else: # self._hyperparameters["findContourPrecision"] == "CHAIN_APPROX_NONE"
      contourPrecision = cv2.CHAIN_APPROX_NONE

    if self._hyperparameters["recalculateForegroundImageBasedOnBodyArea"]:

      minPixel2nbBlackPixels = {}
      countTries = 0
      nbBlackPixels = 0
      nbBlackPixelsMax = int(self._hyperparameters["adjustMinPixelDiffForBackExtract_nbBlackPixelsMax"] / self._hyperparameters["nbAnimalsPerWell"])
      minPixelDiffForBackExtract = int(self._hyperparameters["minPixelDiffForBackExtract"])
      if "minPixelDiffForBackExtractBody" in self._hyperparameters:
        minPixelDiffForBackExtract = self._hyperparameters["minPixelDiffForBackExtractBody"]

      previousNbBlackPixels = []
      while (minPixelDiffForBackExtract > 0) and (countTries < 30) and not(minPixelDiffForBackExtract in minPixel2nbBlackPixels):
        curFrame = initialCurFrame.copy()
        putToWhite = ( curFrame.astype('int32') >= (back.astype('int32') - minPixelDiffForBackExtract) )
        curFrame[putToWhite] = 255
        ret, thresh1_b = cv2.threshold(curFrame, self._hyperparameters["thresholdForBlobImg"], 255, cv2.THRESH_BINARY)
        thresh1_b = 255 - thresh1_b
        bodyContour = 0
        contours, hierarchy = cv2.findContours(thresh1_b, cv2.RETR_TREE, contourPrecision)
        for contour in contours:
          dist = cv2.pointPolygonTest(contour, (x, y), True)
          if dist >= 0:
            M = cv2.moments(contour)
            if M['m00']:
              cx = int(M['m10']/M['m00'])
              cy = int(M['m01']/M['m00'])
              bodyContour = contour
            else:
              cx = 0
              cy = 0
        if not(type(bodyContour) == int):
          nbBlackPixels = cv2.contourArea(bodyContour)
        else:
          nbBlackPixels = -100000000

        minPixel2nbBlackPixels[minPixelDiffForBackExtract] = nbBlackPixels
        if nbBlackPixels > nbBlackPixelsMax:
          minPixelDiffForBackExtract = minPixelDiffForBackExtract + 1
        if nbBlackPixels <= nbBlackPixelsMax:
          minPixelDiffForBackExtract = minPixelDiffForBackExtract - 1

        countTries = countTries + 1

        previousNbBlackPixels.append(nbBlackPixels)
        if len(previousNbBlackPixels) >= 3:
          lastThree = previousNbBlackPixels[len(previousNbBlackPixels)-3: len(previousNbBlackPixels)]
          if lastThree.count(lastThree[0]) == len(lastThree):
            countTries = 1000000

      best_minPixelDiffForBackExtract = 0
      minDist = 10000000000000
      for minPixelDiffForBackExtract in minPixel2nbBlackPixels:
        nbBlackPixels = minPixel2nbBlackPixels[minPixelDiffForBackExtract]
        dist = abs(nbBlackPixels - nbBlackPixelsMax)
        if dist < minDist:
          minDist = dist
          best_minPixelDiffForBackExtract = minPixelDiffForBackExtract

      minPixelDiffForBackExtract = best_minPixelDiffForBackExtract
      putToWhite = (curFrame.astype('int32') >= (back.astype('int32') - minPixelDiffForBackExtract) )
      curFrame[putToWhite] = 255

      ret, thresh1 = cv2.threshold(curFrame, self._hyperparameters["thresholdForBlobImg"], 255, cv2.THRESH_BINARY)
      thresh1 = 255 - thresh1

      self._hyperparameters["minPixelDiffForBackExtractBody"] = minPixelDiffForBackExtract

    contours, hierarchy = cv2.findContours(thresh1, cv2.RETR_TREE, contourPrecision)
    for contour in contours:
      area = cv2.contourArea(contour)
      if (area >= self._hyperparameters["minAreaBody"]) and (area <= self._hyperparameters["maxAreaBody"]):
        dist = cv2.pointPolygonTest(contour, (x, y), True)
        if dist >= 0 or self._hyperparameters["saveBodyMask"]:
          M = cv2.moments(contour)
          if M['m00']:
            cx = int(M['m10']/M['m00'])
            cy = int(M['m01']/M['m00'])
            bodyContour = contour
          else:
            cx = 0
            cy = 0

    if type(bodyContour) != int:
      if cv2.contourArea(bodyContour) >= self._hyperparameters["maxAreaBody"]:
        bodyContour = 0

    if self._hyperparameters["saveBodyMask"]:
      saveImagesAndData(self._hyperparameters, bodyContour, initialCurFrame, wellNumber, frameNumber)

    return bodyContour

  @staticmethod
  def _rotate(boundary, aaa, bbb, angle):
    gauche = 0
    haut   = 0

    x1 = aaa + 100*math.cos(angle)
    y1 = bbb + 100*math.sin(angle)
    x2 = aaa + 100*math.cos(angle + math.pi)
    y2 = bbb + 100*math.sin(angle + math.pi)
    x = 0
    y = 0
    r = 0
    Yoo = [x1 + gauche,y1 + haut]
    Yaa = [x2 + gauche,y2 + haut]

    dist1 = 0
    min_dist1 = 1000000
    dist2 = 0
    min_dist2 = 1000000
    theta = 0
    alpha = 0
    alpha_aux = 0
    final_angle = 0
    for i in range(0, len(boundary)):
      Pt = boundary[i][0]
      dist1 = (Pt[0] - x1)*(Pt[0] - x1) + (Pt[1] - y1)*(Pt[1] - y1)
      dist2 = (Pt[0] - x2)*(Pt[0] - x2) + (Pt[1] - y2)*(Pt[1] - y2)
      if (dist1<min_dist1):
        min_dist1 = dist1
      if (dist2<min_dist2):
        min_dist2 = dist2

    if (min_dist1<min_dist2):
      theta = angle
    else:
      theta = angle + math.pi

    theta = (math.pi/2) - theta

    for i in range(0, len(boundary)):
      Pt = boundary[i][0]
      x = Pt[0]
      y = Pt[1]
      x = x - aaa
      y = y - bbb
      r = math.sqrt(x*x + y*y)
      if (x>0):
        alpha = math.atan(y/x)
      if (x<0):
        x = -x
        alpha_aux = math.atan(y/x)
        alpha = math.pi - alpha_aux
      if (x == 0):
        if (y>0):
          alpha = math.pi/2
        else:
          alpha = -math.pi/2

      final_angle = theta + alpha
      x = r*math.cos(final_angle)
      y = r*math.sin(final_angle)
      Pt[0] = x + aaa
      Pt[1] = y + bbb + 200

      boundary[i] = Pt

    return boundary

  @staticmethod
  def __resampleSeqConstPtsPerArcLength(OrigBound, numTailPoints):
    n = len(OrigBound)
    distOrg = np.zeros(n)
    xOrg    = np.zeros(n)
    yOrg    = np.zeros(n)

    totDist = 0
    distOrg[0] = totDist
    xOrg[0] = OrigBound[0][0][0]
    yOrg[0] = OrigBound[0][0][1]

    for i in range(1, n):
      diff       = math.sqrt((OrigBound[i-1][0][0]-OrigBound[i][0][0])**2 + (OrigBound[i-1][0][1]-OrigBound[i][0][1])**2)
      totDist    = totDist + diff
      distOrg[i] = totDist

    uniDist = np.zeros(numTailPoints)
    uniX    = np.zeros(numTailPoints)
    uniY    = np.zeros(numTailPoints)

    for i in range(0, numTailPoints):
      uniDist[i] = totDist * (i/(numTailPoints-1))

    for i in range(1, n):
      xOrg[i] = OrigBound[i][0][0]
      yOrg[i] = OrigBound[i][0][1]

    uniX = np.interp(uniDist, distOrg, xOrg)
    uniY = np.interp(uniDist, distOrg, yOrg)

    output = np.zeros((numTailPoints, 2))
    for i in range(0, numTailPoints):
      output[i][0] = uniX[i]
      output[i][1] = uniY[i]

    return output

  @staticmethod
  def __fillTailRanges(tailRange1,tailRange2,fillSecond,i,MostCurvyIndex):
    if (i == MostCurvyIndex):
      fillSecond = 1
    if fillSecond == 0:
      tailRange1.append(i)
    else:
      tailRange2.append(i)
    return [tailRange1,tailRange2,fillSecond]

  def _getMidline(self, bord1, bord2, MostCurvyIndex, boundary, dst, nbTailPoints, distance2, debug):
    output = np.zeros((1, 0, 2))

    minTailSize = 20
    maxTailSize = 60
    trackingPointSizeDisplay = 1

    OrigBoundA = []
    OrigBoundB = []

    if (bord2 < bord1):
      temp  = bord2
      bord2 = bord1
      bord1 = temp

    max1 = distance2[bord2] - distance2[bord1]
    max2 = (distance2[bord1] - distance2[0])  + (distance2[len(boundary)] - distance2[bord2])

    tailRangeA = []
    tailRangeB = []
    fillSecond = 0
    if (max1 > max2):
      for i in range(bord1, bord2):
        [tailRangeA,tailRangeB,fillSecond] = self.__fillTailRanges(tailRangeA,tailRangeB,fillSecond,i,MostCurvyIndex)
    else:
      for i in range(bord2, len(boundary)):
        [tailRangeA,tailRangeB,fillSecond] = self.__fillTailRanges(tailRangeA,tailRangeB,fillSecond,i,MostCurvyIndex)
      for i in range(0, bord1):
        [tailRangeA,tailRangeB,fillSecond] = self.__fillTailRanges(tailRangeA,tailRangeB,fillSecond,i,MostCurvyIndex)

    OrigBoundA = boundary[tailRangeA]
    OrigBoundB = boundary[tailRangeB]

    if ((bord1!=bord2) and (bord1!=MostCurvyIndex) and (bord2!=MostCurvyIndex) and not((bord1==1) and (bord2==1) and (MostCurvyIndex==1)) and (len(OrigBoundA)>1) and (len(OrigBoundB)>1)):

      if False:
        for pt in OrigBoundA:
          cv2.circle(dst, (pt[0][0], pt[0][1]), 1, (0, 255, 0), -1)
        for pt in OrigBoundB:
          cv2.circle(dst, (pt[0][0], pt[0][1]), 1, (255, 0, 0), -1)
        self._debugFrame(dst, title='dst')

      NBoundA = self.__resampleSeqConstPtsPerArcLength(OrigBoundA, nbTailPoints)
      NBoundB = self.__resampleSeqConstPtsPerArcLength(OrigBoundB, nbTailPoints)

      # calculates length of the tail
      TotalDist = 0
      for i in range(1, nbTailPoints):
        Pt  = NBoundB[i % nbTailPoints]
        Pt2 = NBoundA[nbTailPoints - i]
        x = (Pt[0]+Pt2[0]) / 2
        y = (Pt[1]+Pt2[1]) / 2
        if i > 1:
          TotalDist = TotalDist + math.sqrt((x-xAvant)*(x-xAvant)+(y-yAvant)*(y-yAvant))
        xAvant = x
        yAvant = y

      if ((TotalDist<self._hyperparameters["minTailSize"]) or (TotalDist>self._hyperparameters["maxTailSize"])):

        if (debug):
          print("innapropriate tail size! TailDist: ", TotalDist, " ; but minTailSize is ", minTailSize, " and maxTailSize is ", maxTailSize)

      else:

        Tail = boundary[MostCurvyIndex][0]

        point = np.array([Tail[0], Tail[1]])
        output = np.insert(output, 0, point, axis=1)

        for i in range(1, nbTailPoints):
          Pt  = NBoundB[i % nbTailPoints]
          Pt2 = NBoundA[nbTailPoints - i]
          point = np.array([(Pt[0]+Pt2[0])/2, (Pt[1]+Pt2[1])/2])
          output = np.insert(output, 0, point, axis=1)

        i = nbTailPoints-2
        if i >= 1:
          Pt =  NBoundB[i % nbTailPoints]
          Pt2 = NBoundA[nbTailPoints-i]
          ClosestPoint = [ (Pt[0]+Pt2[0])/2 , (Pt[1]+Pt2[1])/2 ]
        else:
          ClosestPoint = [-200, -200]

    else:

      # THIS SHOULD BE IMPROVED IN THE FUTURE:
      # WE SHOULD CHECK FOR TAIL LENGHT
      # ALSO WE SHOULD DO SOMETHING BETTER THAN JUST PUTTING THE TAIL TIP FOR EACH OF THE TEN POINTS !!!
      Tail = boundary[MostCurvyIndex][0]
      point = np.array([Tail[0], Tail[1]])
      for i in range(0, nbTailPoints):
        output = np.insert(output, 0, point, axis=1)

    return output

  def _tailTrackingExtremityDetect(self, headPosition, i, thresh1, frame, debugAdv, heading, initialCurFrame, back, wellNumber=-1):
    newHeading = -1

    dst = frame.copy()
    if type(dst[0][0]) == np.uint8:
      dst = cv2.cvtColor(dst, cv2.COLOR_GRAY2RGB)
    firstFrame = self._hyperparameters["firstFrame"]
    lastFrame = self._hyperparameters["lastFrame"]

    if self._hyperparameters["debugTrackingThreshImg"]:
      if self._hyperparameters["debugTrackingPtExtremeLargeVerticals"]:
        self._debugFrame(thresh1[int(headPosition[1])-200:len(thresh1), :], title='debugTrackingThreshImg')
      else:
        self._debugFrame(thresh1, title='debugTrackingThreshImg')

    # Finding blob corresponding to the body of the fish
    bodyContour = self.__findBodyContour(headPosition, thresh1, initialCurFrame, back, wellNumber, i)
    if type(bodyContour) != int:
      # Finding the two sides of the fish
      res = self.__findTheTwoSides(headPosition, bodyContour, dst)
      if len(res) == 3:
        heading = res[2]
        newHeading = res[2]
      # Finding tail extremity
      rotatedContour = bodyContour.copy()
      rotatedContour = self._rotate(rotatedContour,int(headPosition[0]),int(headPosition[1]),heading)
      [MostCurvyIndex, distance2] = self._findTailExtremete(rotatedContour, bodyContour, headPosition[0], int(res[0]), int(res[1]), debugAdv, dst, self._hyperparameters["tailExtremityMaxJugeDecreaseCoeff"])
      if debugAdv:
        if True:
          # Head Center
          cv2.circle(dst, (int(headPosition[0]),int(headPosition[1])), 3, (255, 255, 0), -1)
          # Tail basis 1
          pt1 = bodyContour[int(res[0])][0]
          cv2.circle(dst, (pt1[0],pt1[1]), 3, (255, 0, 0), -1)
          # Tail basis 2
          pt1 = bodyContour[int(res[1])][0]
          cv2.circle(dst, (pt1[0],pt1[1]), 3, (180, 0, 0), -1)
          # Tail extremity
          pt1 = bodyContour[int(MostCurvyIndex)][0]
          cv2.circle(dst, (pt1[0],pt1[1]), 3, (0, 0, 255), -1)
        else:
          for pt in bodyContour:
            cv2.circle(dst, (pt[0][0], pt[0][1]), 1, (0, 255, 0), -1)
          cv2.circle(dst, (int(headPosition[0]),int(headPosition[1])), 1, (0, 0, 255), -1)
        #
        if self._hyperparameters["debugTrackingPtExtremeLargeVerticals"]:
          dst = dst[int(headPosition[1])-200:len(dst), :]
        # Plotting points
        self._debugFrame(dst, title='Frame')

      # Getting Midline
      if self._hyperparameters["detectMouthInsteadOfHeadTwoSides"] == 0:
        tail = self._getMidline(int(res[0]), int(res[1]), int(MostCurvyIndex), bodyContour, dst, self._nbTailPoints-1, distance2, debugAdv)
      else:
        tail = self._getMidline(int(res[0]), int(res[1]), int(MostCurvyIndex), bodyContour, dst, self._nbTailPoints, distance2, debugAdv)
        tail = np.array([tail[0][1:len(tail[0])]])

      if False:
        maxDistContourToTail = -1
        for contourPt in bodyContour:
          contourPtX = contourPt[0][0]
          contourPtY = contourPt[0][1]
          minDistContourPointToTail = 1000000000000
          for tailPt in np.append(tail[0], np.array([headPosition]), axis=0):
            tailPtX = tailPt[0]
            tailPtY = tailPt[1]
            dist = math.sqrt((tailPtX - contourPtX)**2 + (tailPtY - contourPtY)**2)
            if dist < minDistContourPointToTail:
              minDistContourPointToTail = dist
          if minDistContourPointToTail > maxDistContourToTail:
            maxDistContourToTail = minDistContourPointToTail
        print("maxDistContourToTail:", maxDistContourToTail, "; tailLength:", self._hyperparameters["minTailSize"]*10)

      if False:
        for pt in bodyContour:
          cv2.circle(dst, (pt[0][0], pt[0][1]), 3, (0, 0, 255), -1)
        self._debugFrame(dst, title='Frame')

      # Optimizing midline if necessary
      midlineIsInBlobTrackingOptimization = self._hyperparameters["midlineIsInBlobTrackingOptimization"]
      if midlineIsInBlobTrackingOptimization:
        [allInside, tailLength] = self.__checkIfMidlineIsInBlob(int(res[0]), int(res[1]), int(MostCurvyIndex), bodyContour, dst, self._nbTailPoints-1, distance2, debugAdv)
        if allInside == False:
          n = len(bodyContour)
          maxTailLength = -1
          for j in range(0, n):
            [allInside, tailLength] = self.__checkIfMidlineIsInBlob(int(res[0]), int(res[1]), j, bodyContour, dst, self._nbTailPoints-1, distance2, debugAdv)
            if allInside:
              if tailLength > maxTailLength:
                MostCurvyIndex = j
                maxTailLength = tailLength
          tail = self._getMidline(int(res[0]), int(res[1]), int(MostCurvyIndex), bodyContour, dst, self._nbTailPoints-1, distance2, debugAdv)
      # Applying snake on tail
      applySnake = False
      if applySnake:
        tail2 = tail[0]
        n = len(tail2)
        # tail2[n-1][0] = tail2[n-1][0] + (tail2[n-1][0] - tail2[n-2][0]) * 6
        # tail2[n-1][1] = tail2[n-1][1] + (tail2[n-1][1] - tail2[n-2][1]) * 6
        # print(type(tail))
        # r = np.linspace(tail2[0][0], tail2[0][0] + (tail2[1][0]-tail2[0][0]) * 15, 9)
        # c = np.linspace(tail2[0][1], tail2[0][1] + (tail2[1][1]-tail2[0][1]) * 15, 9)
        # tail2 = np.array([r, c]).T
        # r = np.linspace(tail2[0][0], tail2[n-1][0], 9)
        # c = np.linspace(tail2[0][1], tail2[n-1][1], 9)
        # tail2 = np.array([r, c]).T
        from skimage.color import rgb2gray
        from skimage.filters import gaussian
        from skimage.segmentation import active_contour
        snake = active_contour(gaussian(frame, 3), tail2, w_edge=-1000, bc="fixed")
        # snake = active_contour(gaussian(frame, 3), tail2, w_edge=0, bc="fixed-free")
        print(snake)
        # snake = tail2
        tail[0] = snake

    else:

      tail = np.zeros((1, 0, 2))
      point = np.array([0, 0])
      for i in range(0, self._nbTailPoints):
        tail = np.insert(tail, 0, point, axis=1)
      # if self._hyperparameters["detectMouthInsteadOfHeadTwoSides"] != 0:
        # tail = np.insert(tail, 0, point, axis=1)

    # Inserting head position, smoothing tail and creating output
    # if self._hyperparameters["detectMouthInsteadOfHeadTwoSides"] == 0:
      # tail = np.insert(tail, 0, headPosition, axis=1)
    tail = np.insert(tail, 0, headPosition, axis=1)

    # if self._nbTailPoints != len(tail[0]):
      # print("small problem 1 in tailTrackingExtremityDetect")

    # output = np.zeros((1, len(tail[0]), 2))
    output = np.zeros((1, self._nbTailPoints, 2))

    for idx, x in enumerate(tail[0]):
      if idx < self._nbTailPoints:
        output[0][idx][0] = x[0]
        output[0][idx][1] = x[1]
      # else:
        # print("small problem 2 in tailTrackingExtremityDetect")

    return [output, newHeading]



import numpy as np
from scipy.interpolate import UnivariateSpline

def getTailAngleRecalculated(curbout, nbFramesTakenIntoAccount, numberOfBendsIncludedForMaxDetect):

  fin = curbout["BoutEnd"] - curbout["BoutStart"] + 1

  numps = 3
  
  headx = curbout["HeadX"]
  heady = curbout["HeadY"]
  tailx = curbout["TailX_VideoReferential"]
  taily = curbout["TailY_VideoReferential"]
  
  tailangles_arr = np.zeros(fin)
  for i in range(0, fin):
    if len(taily[i]) > 3 and len(tailx[i]) > 3:
      ang = np.arctan2(heady[i] - taily[i][-3], headx[i] - tailx[i][-3])*180/np.pi
      ang2 = np.arctan2(heady[i] - taily[i][0], headx[i] - tailx[i][0])*180/np.pi
      delang = ang2 - ang
      if np.abs(delang) < 180:
        tailangles_arr[i] = delang
      elif delang > 180:
        tailangles_arr[i] = delang - 360
      elif delang < -180:
        tailangles_arr[i] = 360 + delang
    else:
      tailangles_arr[i] = 0
      # base = min(len(taily[i]), len(tailx[i])) - 1
    # if i==0:
      # print("yo:")
      # print(heady[i])
      # print(taily[i])
      # print(headx[i])
      # print(tailx[i])
    #print(i,j,ang,ang2,tailangles_arr[i,j])
  
  tailAngle3 = [0]*nbFramesTakenIntoAccount
  sizeTabNotZero = min(fin, nbFramesTakenIntoAccount)
  tailAngle3[0:sizeTabNotZero]  = tailangles_arr[0:sizeTabNotZero]
  
  # Is this necessary ???
  # Bend_Amplitude = curbout["Bend_Amplitude"].copy()
  # Bend_Amplitude.insert(0, 0)
  # Bend_Amplitude.append(0)
  # if numberOfBendsIncludedForMaxDetect == -1 or numberOfBendsIncludedForMaxDetect >= len(Bend_Amplitude):
    # argmax = np.argmax(np.abs(Bend_Amplitude))
  # else:
    # Bend_Amplitude_Array = Bend_Amplitude[0:numberOfBendsIncludedForMaxDetect]
    # argmax               = np.argmax(np.abs(Bend_Amplitude_Array))
  # maxAbsBendValue = Bend_Amplitude[argmax]
  # tailAngle = 0
  # if maxAbsBendValue < 0:
    # tailAngle = (-np.array(curbout["TailAngle_smoothed"])).tolist()
    # instMean  = (-np.array(instMean)).tolist()
  # else:
    # tailAngle = curbout["TailAngle_smoothed"]
  
  return tailAngle3



import numpy as np
import cv2

import zebrazoom.videoFormatConversion.zzVideoReading as zzVideoReading
from zebrazoom.code.preprocessImage import preprocessBackgroundImage


class GetBackgroundMixin:
  def _getBackground(self):
    cap   = zzVideoReading.VideoCapture(self._videoPath)
    max_l = int(cap.get(7))

    backCalculationStep = self._hyperparameters["backCalculationStep"]
    if ("firstFrameForBackExtract" in self._hyperparameters) and ("lastFrameForBackExtract" in self._hyperparameters) and self._hyperparameters["firstFrameForBackExtract"] != -1 and self._hyperparameters["lastFrameForBackExtract"] != -1:
      firstFrame = self._hyperparameters["firstFrameForBackExtract"]
      lastFrame  = self._hyperparameters["lastFrameForBackExtract"]
    else:
      if ("firstFrame" in self._hyperparameters) and (self._hyperparameters["backgroundExtractionForceUseAllVideoFrames"] == 0):
        firstFrame = self._hyperparameters["firstFrame"]
      else:
        firstFrame = 1
      if ("lastFrame" in self._hyperparameters) and (self._hyperparameters["backgroundExtractionForceUseAllVideoFrames"] == 0):
        lastFrame  = self._hyperparameters["lastFrame"]
      else:
        lastFrame  = max_l - 10

    debugExtractBack    = self._hyperparameters["debugExtractBack"]

    if backCalculationStep == -1:
      backCalculationStep = int((lastFrame - firstFrame) / self._hyperparameters["nbImagesForBackgroundCalculation"])
      if backCalculationStep <= 1:
        backCalculationStep = 1

    ret, back = cap.read()

    if self._hyperparameters["useFirstFrameAsBackground"]:
      if self._hyperparameters["invertBlackWhiteOnImages"]:
        back = 255 - back
      back = cv2.cvtColor(back, cv2.COLOR_BGR2GRAY)
      if self._hyperparameters["backgroundPreProcessMethod"]:
        back = preprocessBackgroundImage(back, self._hyperparameters)
      if debugExtractBack:
        self._debugFrame(back, title="Background Extracted", timeout = 3000 if self._hyperparameters["exitAfterBackgroundExtraction"] else None)
      cap.release()
      print("Background Extracted from first frame")
      return back

    if ret and self._hyperparameters["invertBlackWhiteOnImages"]:
      back = 255 - back
    back = cv2.cvtColor(back, cv2.COLOR_BGR2GRAY)
    if self._hyperparameters["backgroundExtractionWithOnlyTwoFrames"] == 0:
      for k in range(firstFrame,lastFrame):
        if (k % backCalculationStep == 0):
          cap.set(1, k)
          ret, frame = cap.read()
          if ret and self._hyperparameters["invertBlackWhiteOnImages"]:
            frame = 255 - frame
          if debugExtractBack:
            print(k)
          if ret:
            frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
            if self._hyperparameters["extractBackWhiteBackground"]:
              back = cv2.max(frame, back)
            else:
              back = cv2.min(frame, back)
          else:
            print("couldn't use the frame", k, "for the background extraction")
    else:
      maxDiff    = 0
      indMaxDiff = firstFrame
      for k in range(firstFrame,lastFrame):
        if (k % backCalculationStep == 0):
          cap.set(1, k)
          ret, frame = cap.read()
          if ret:
            if self._hyperparameters["invertBlackWhiteOnImages"]:
              frame = 255 - frame
            frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
            diff  = np.sum(np.abs(frame - back))
            if diff > maxDiff:
              maxDiff    = diff
              indMaxDiff = k
      cap.set(1, indMaxDiff)
      ret, frame = cap.read()
      if ret:
        if self._hyperparameters["invertBlackWhiteOnImages"]:
          frame = 255 - frame
        frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
        if self._hyperparameters["extractBackWhiteBackground"]:
          back = cv2.max(frame, back)
        else:
          back = cv2.min(frame, back)
      else:
        print("couldn't use the frame", k, "for the background extraction")

    if self._hyperparameters["backgroundPreProcessMethod"]:
      back = preprocessBackgroundImage(back, self._hyperparameters)

    if self._hyperparameters["checkThatMovementOccurInVideo"]:
      cap.set(1, firstFrame)
      ret, frame = cap.read()
      if ret:
        if self._hyperparameters["invertBlackWhiteOnImages"]:
          frame = 255 - frame
        frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
        if self._hyperparameters["imagePreProcessMethod"]:
          frame = preprocessBackgroundImage(frame, self._hyperparameters)
        if type(frame[0][0]) == np.ndarray:
          frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
        putToWhite = (frame.astype('int32') >= (back.astype('int32')-self._hyperparameters["minPixelDiffForBackExtract"]))
        frame[putToWhite] = 255
      firstImage = frame
      maxDiff    = 0
      indMaxDiff = firstFrame
      for k in range(firstFrame,lastFrame):
        if (k % backCalculationStep == 0):
          cap.set(1, k)
          ret, frame = cap.read()
          if ret:
            if self._hyperparameters["invertBlackWhiteOnImages"]:
              frame = 255 - frame
            frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
            if self._hyperparameters["imagePreProcessMethod"]:
              frame = preprocessBackgroundImage(frame, self._hyperparameters)
            if type(frame[0][0]) == np.ndarray:
              frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
            putToWhite = (frame.astype('int32') >= (back.astype('int32')-self._hyperparameters["minPixelDiffForBackExtract"]))
            frame[putToWhite] = 255
            if self._hyperparameters["checkThatMovementOccurInVideoMedianFilterWindow"]:
              diff  = np.sum(cv2.medianBlur(np.abs(frame - firstImage), self._hyperparameters["checkThatMovementOccurInVideoMedianFilterWindow"]))
            else:
              diff  = np.sum(np.abs(frame - firstImage))
            if diff > maxDiff:
              maxDiff    = diff
              indMaxDiff = k
      print("checkThatMovementOccurInVideo: max difference is:", maxDiff)
      if maxDiff < self._hyperparameters["checkThatMovementOccurInVideo"]:
        back[:, :] = 0 # TODO: tracking should NOT RUN after background is set to 0 as it is here

    if self._hyperparameters["setBackgroundToImageMedian"]:
      back[:, :] = np.median(back)

    if debugExtractBack:
      self._debugFrame(back, title="Background Extracted", timeout = 3000 if self._hyperparameters["exitAfterBackgroundExtraction"] else None)
    cap.release()

    print("Background Extracted")
    if self._hyperparameters["popUpAlgoFollow"]:
      import zebrazoom.code.popUpAlgoFollow as popUpAlgoFollow
      popUpAlgoFollow.prepend("Background Extracted")

    if self._hyperparameters["exitAfterBackgroundExtraction"]:
      print("exitAfterBackgroundExtraction")
      raise ValueError

    return back


def getBackground(videoPath, hyperparameters):
  obj = GetBackgroundMixin()
  obj._videoPath = videoPath
  obj._hyperparameters = hyperparameters
  return obj._getBackground()



import shutil

import zebrazoom.code.paths as paths
from zebrazoom.dataAnalysis.datasetcreation.createDataFrame import createDataFrame
from zebrazoom.dataAnalysis.dataanalysis.populationComparaison import populationComparaison
from zebrazoom.dataAnalysis.postProcessingFromCommandLine.postProcessingFromCommandLine import calculateNumberOfSfsVsTurnsBasedOnMaxAmplitudeThreshold
from zebrazoom.dataAnalysis.datasetcreation.generatePklDataFileForVideo import generatePklDataFileForVideo
import os

def kinematicParametersAnalysis(sys, addMedianPerGenotype=0, checkConsistencyOfParameters=False):
  
  pathToExcelFile                 = sys.argv[3]
  
  frameStepForDistanceCalculation = int(sys.argv[4]) if len(sys.argv) >= 5 else 4
  
  if len(sys.argv) >= 7:
    minNbBendForBoutDetect = int(sys.argv[5])
    keep                   = int(sys.argv[6])
  else:
    minNbBendForBoutDetect = -1
    keep                   = 1
  
  angleThreshSFSvsTurns  = int(sys.argv[7]) if len(sys.argv) >= 8 else -1
  
  tailAngleKinematicParameterCalculation    = int(sys.argv[8]) if len(sys.argv) >= 9 else 1
  
  saveRawDataInAllBoutsSuperStructure       = int(sys.argv[9]) if len(sys.argv) >= 10 else 1
  
  saveAllBoutsSuperStructuresInMatlabFormat = int(sys.argv[10]) if len(sys.argv) >= 11 else 1
  
  forcePandasDfRecreation                   = int(sys.argv[11]) if len(sys.argv) >= 12 else 0
  
  nameWithExt = os.path.split(pathToExcelFile)[1]

  dataframeOptions = {
    'pathToExcelFile'                   : os.path.split(pathToExcelFile)[0],
    'fileExtension'                     : os.path.splitext(nameWithExt)[1],
    'resFolder'                         : os.path.join(paths.getDataAnalysisFolder(), 'data'),
    'nameOfFile'                        : os.path.splitext(nameWithExt)[0],
    'smoothingFactorDynaParam'          : 0,
    'nbFramesTakenIntoAccount'          : 0,
    'numberOfBendsIncludedForMaxDetect' : -1,
    'minNbBendForBoutDetect'            : minNbBendForBoutDetect,
    'keepSpeedDistDurWhenLowNbBends'    : keep,
    'defaultZZoutputFolderPath'         : paths.getDefaultZZoutputFolder(),
    'computeTailAngleParamForCluster'   : False,
    'computeMassCenterParamForCluster'  : False,
    'frameStepForDistanceCalculation'   : str(frameStepForDistanceCalculation),
    'tailAngleKinematicParameterCalculation'    : tailAngleKinematicParameterCalculation,
    'saveRawDataInAllBoutsSuperStructure'       : saveRawDataInAllBoutsSuperStructure,
    'saveAllBoutsSuperStructuresInMatlabFormat' : saveAllBoutsSuperStructuresInMatlabFormat
  }
  
  generatePklDataFileForVideo(os.path.join(os.path.split(pathToExcelFile)[0], nameWithExt), paths.getDefaultZZoutputFolder(), frameStepForDistanceCalculation, forcePandasDfRecreation)
  
  [conditions, genotypes, nbFramesTakenIntoAccount, globParam] = createDataFrame(dataframeOptions, '', forcePandasDfRecreation, [])

  outputFolder = os.path.join(paths.getDataAnalysisFolder(), 'resultsKinematic')
  resultFolder = os.path.join(outputFolder, dataframeOptions['nameOfFile'])
  if os.path.exists(resultFolder):
    shutil.rmtree(resultFolder)  # if the result folder exists, remove it manually, since populationComparaison only removes it if plotOutliersAndMean argument is True

  outliersRemoved = minNbBendForBoutDetect
  # Mixing up all the bouts
  if not outliersRemoved:  # check if outliers are already removed from results
    populationComparaison(dataframeOptions['nameOfFile'], dataframeOptions['resFolder'], globParam, conditions, genotypes, outputFolder, 0, True, 0, 0, int(checkConsistencyOfParameters))
  
  if not(checkConsistencyOfParameters):
    populationComparaison(dataframeOptions['nameOfFile'], dataframeOptions['resFolder'], globParam, conditions, genotypes, outputFolder, 0, False, 0, 0, int(checkConsistencyOfParameters))
  
  # Median per well for each kinematic parameter
  if not outliersRemoved:  # check if outliers are already removed from results
    populationComparaison(dataframeOptions['nameOfFile'], dataframeOptions['resFolder'], globParam, conditions, genotypes, outputFolder, 1, True, 0, 0, int(checkConsistencyOfParameters))
  
  if not(checkConsistencyOfParameters):
    populationComparaison(dataframeOptions['nameOfFile'], dataframeOptions['resFolder'], globParam, conditions, genotypes, outputFolder, 1, False, 0, 0, int(checkConsistencyOfParameters))
  
  # Median per genotype for each kinematic parameter
  if addMedianPerGenotype:
    populationComparaison(dataframeOptions['nameOfFile'], dataframeOptions['resFolder'], globParam, conditions, genotypes, outputFolder, 0, True, 0, 1, int(checkConsistencyOfParameters))
    
    if not(checkConsistencyOfParameters):
      populationComparaison(dataframeOptions['nameOfFile'], dataframeOptions['resFolder'], globParam, conditions, genotypes, outputFolder, 0, False, 0, 1, int(checkConsistencyOfParameters))
  
  if angleThreshSFSvsTurns != -1:
    calculateNumberOfSfsVsTurnsBasedOnMaxAmplitudeThreshold(paths.getRootDataFolder(), dataframeOptions['nameOfFile'], angleThreshSFSvsTurns)
  
  print("")
  print("")
  print("The data has been saved in the folder:", resultFolder)
  print("The raw data has also been saved in:", dataframeOptions['resFolder'])



import os
import scipy
import scipy.io
import pandas as pd
import json
import numpy as np
import warnings
warnings.filterwarnings('ignore', category=np.VisibleDeprecationWarning)
from zebrazoom.dataAnalysis.datasetcreation.getDynamicParameters import getDynamicParameters
from zebrazoom.dataAnalysis.datasetcreation.getTailAngles import getTailAngles
from zebrazoom.dataAnalysis.datasetcreation.getInstaSpeed import getInstaSpeed
from zebrazoom.dataAnalysis.datasetcreation.getInstaHeadingDiff import getInstaHeadingDiff
from zebrazoom.dataAnalysis.datasetcreation.getInstaHorizontalDisplacement import getInstaHorizontalDisplacement
from zebrazoom.dataAnalysis.datasetcreation.getGlobalParameters import getGlobalParameters
from zebrazoom.dataAnalysis.datasetcreation.getTailLength  import getTailLength
from zebrazoom.dataAnalysis.datasetcreation.getTailLength2 import getTailLength2
from zebrazoom.dataAnalysis.datasetcreation.getTailAngleRecalculated import getTailAngleRecalculated
from zebrazoom.dataAnalysis.datasetcreation.getTailAngleRecalculated2 import getTailAngleRecalculated2
from zebrazoom.dataAnalysis.datasetcreation.gatherInitialRawData import gatherInitialRawData
import pickle

def createDataFrame(dataframeOptions, excelFileDataFrame="", forcePandasDfRecreation=0, addToGlobalParameters=0, minimumFrameToFrameDistanceToBeConsideredAsMoving=0, supstructOverwrite={}):

  # Gathering user-inputed information about how to create the dataframe of parameters for the whole set of videos
  
  pathToExcelFile                   = dataframeOptions['pathToExcelFile']
  nameOfFile                        = dataframeOptions['nameOfFile']
  fileExtension                     = dataframeOptions['fileExtension']
  smoothingFactor                   = dataframeOptions['smoothingFactorDynaParam']
  resFolder                         = dataframeOptions['resFolder']
  numberOfBendsIncludedForMaxDetect = dataframeOptions['numberOfBendsIncludedForMaxDetect']
  minNbBendForBoutDetect            = dataframeOptions['minNbBendForBoutDetect']
  computeTailAngleParamForCluster   = dataframeOptions['computeTailAngleParamForCluster']
  computeMassCenterParamForCluster  = dataframeOptions['computeMassCenterParamForCluster']
  
  defaultZZoutputFolderPath = dataframeOptions['defaultZZoutputFolderPath'] if 'defaultZZoutputFolderPath' in dataframeOptions else ''
  
  computetailAnglesRecalculatedParamsForCluster = dataframeOptions["computetailAnglesRecalculatedParamsForCluster"] if 'computetailAnglesRecalculatedParamsForCluster' in dataframeOptions else False
  
  keepSpeedDistDurWhenLowNbBends = int(dataframeOptions['keepSpeedDistDurWhenLowNbBends']) if 'keepSpeedDistDurWhenLowNbBends' in dataframeOptions else 1
  
  frameStepForDistanceCalculation = int(dataframeOptions['frameStepForDistanceCalculation']) if ('frameStepForDistanceCalculation' in dataframeOptions) and len(dataframeOptions['frameStepForDistanceCalculation']) else 4
  
  tailAngleKinematicParameterCalculation = int(dataframeOptions['tailAngleKinematicParameterCalculation']) if 'tailAngleKinematicParameterCalculation' in dataframeOptions else 0
  
  saveRawDataInAllBoutsSuperStructure = int(dataframeOptions['saveRawDataInAllBoutsSuperStructure']) if 'saveRawDataInAllBoutsSuperStructure' in dataframeOptions else 0
  
  saveAllBoutsSuperStructuresInMatlabFormat = int(dataframeOptions['saveAllBoutsSuperStructuresInMatlabFormat']) if 'saveAllBoutsSuperStructuresInMatlabFormat' in dataframeOptions else 0
  
  getTailAngleSignMultNormalized = int(dataframeOptions['getTailAngleSignMultNormalized']) if 'getTailAngleSignMultNormalized' in dataframeOptions else 0

  gaussianFitOutlierRemoval = dataframeOptions.get('gaussianFitOutlierRemoval', False)

  # If nbFramesTakenIntoAccount was not specified, finds an appropriate value for it
  nbFramesTakenIntoAccount          = dataframeOptions['nbFramesTakenIntoAccount']
  if len(pathToExcelFile):
    excelFile = pd.read_excel(os.path.join(pathToExcelFile, nameOfFile + fileExtension))
  elif type(excelFileDataFrame) != str:
    excelFile = excelFileDataFrame
  else:
    print("You must provide either an excel file or a video name to create a dataframe of parameters.")
  
  if nbFramesTakenIntoAccount <= -1:
    boutNbFrames = []
    boutTakenIntoAcccount = 0
    videoId = 0
    if excelFile.loc[videoId, 'path'] == "defaultZZoutputFolder":
      path    = os.path.join(defaultZZoutputFolderPath, excelFile.loc[videoId, 'trial_id'])
    else:
      path    = os.path.join(excelFile.loc[videoId, 'path'], excelFile.loc[videoId, 'trial_id'])
    trial_id  = excelFile.loc[videoId, 'trial_id']
    include   = [bool(int(val.strip())) for val in excelFile.loc[videoId, 'include'][1:-1].split(',')]
    if len(supstructOverwrite):
      supstruct = supstructOverwrite
    else:
      with open(os.path.join(path, 'results_' + trial_id + '.txt')) as f:
        supstruct = json.load(f)
    for Well_ID, Cond in enumerate(include):
      if include[Well_ID]:
        for NumBout, dataForBout in enumerate(supstruct["wellPoissMouv"][Well_ID][0]):
          boutNbFrames.append(len(dataForBout["HeadX"]))
          boutTakenIntoAcccount = boutTakenIntoAcccount + 1
          if boutTakenIntoAcccount > 100:
            break;
      if boutTakenIntoAcccount > 100:
        break;
    if len(boutNbFrames):
      if nbFramesTakenIntoAccount == -1:
        nbFramesTakenIntoAccount = int(np.median(boutNbFrames))
      else:
        nbFramesTakenIntoAccount = 3 * int(np.median(boutNbFrames)) if 3 * int(np.median(boutNbFrames)) >= 100 else 100
    else:
      # There are not bouts in the dataset...
      return [[], [], 0, []]
  
  # Creating labels of columns of dataframe
  # General basic information
  basicInformation = ['Trial_ID', 'Well_ID', 'NumBout', 'BoutStart', 'BoutEnd', 'Condition', 'Genotype', 'videoDuration']
  # Global parameters
  if tailAngleKinematicParameterCalculation:
    globParam  = ['Bout Duration (s)', 'Bout Distance (mm)', 'Bout Speed (mm/s)', 'Angular Velocity (deg/s)', 'Max TBF (Hz)', 'Mean TBF (Hz)', 'medianOfInstantaneousTBF', 'Max absolute TBA (deg.)', 'maxBendAmplitudeSigned', 'Mean absolute TBA (deg.)', 'Median absolute TBA (deg.)', 'medianBendAmplitudeSigned', 'Number of Oscillations', 'meanTBF', 'maxTailAngleAmplitude', 'Absolute Yaw (deg)', 'Signed Yaw (deg)', 'TBA#1 timing (s)', 'TBA#1 Amplitude (deg)', 'firstBendAmplitudeSigned', 'IBI (s)', 'xmean', 'ymean', 'binaryClass25degMaxTailAngle', 'tailAngleIntegralSigned', 'BoutFrameNumberStart', 'tailAngleSymmetry', 'secondBendAmpDividedByFirst', 'tailAngleIntegral', 'maxInstantaneousSpeed']
  else:
    globParam  = ['Bout Duration (s)', 'Bout Distance (mm)', 'Bout Speed (mm/s)', 'Angular Velocity (deg/s)']
  
  if type(addToGlobalParameters) == list:
    globParam = globParam + addToGlobalParameters  
  
  # Initial raw data
  if saveRawDataInAllBoutsSuperStructure:
    if tailAngleKinematicParameterCalculation:
      rawData = ['HeadX', 'HeadY', 'Heading', 'TailAngle_Raw', 'TailAngle_smoothed', 'Bend_Timing', 'Bend_TimingAbsolute', 'Bend_Amplitude', 'TailBeatFrequency', 'curvature']
    else:
      rawData = ['HeadX', 'HeadY', 'Heading']
  # Tail angle related parameters for clustering
  instaTBF   = ['instaTBF'  + str(i) for i in range(1,nbFramesTakenIntoAccount+1)]
  instaAmp   = ['instaAmp'  + str(i) for i in range(1,nbFramesTakenIntoAccount+1)]
  instaAsym  = ['instaAsym' + str(i) for i in range(1,nbFramesTakenIntoAccount+1)]
  tailAngles = ['tailAngles'+ str(i) for i in range(1,nbFramesTakenIntoAccount+1)]
  tailAnglesRecalculated = ['tailAnglesRecalculated'+ str(i) for i in range(1,nbFramesTakenIntoAccount+1)]
  tailAnglesRecalculated2 = ['tailAnglesRecalculated2_'+ str(i) for i in range(1, 7*nbFramesTakenIntoAccount+1)]
  # Center of mass related parameters for clustering
  instaSpeed       = ['instaSpeed'       + str(i) for i in range(1,nbFramesTakenIntoAccount+1)]
  instaHeadingDiff = ['instaHeadingDiff' + str(i) for i in range(1,nbFramesTakenIntoAccount+1)]
  instaHorizDispl  = ['instaHorizDispl'  + str(i) for i in range(1,nbFramesTakenIntoAccount+1)]
  # Assembling columns
  dfCols = basicInformation + globParam
  if computeTailAngleParamForCluster:
    dfCols = dfCols + tailAngles
  if saveRawDataInAllBoutsSuperStructure:
    dfCols = dfCols + rawData
  if computeTailAngleParamForCluster:
    dfCols = dfCols + instaTBF + instaAmp + instaAsym
  if computeMassCenterParamForCluster:
    dfCols = dfCols + instaSpeed + instaHeadingDiff + instaHorizDispl
  if computetailAnglesRecalculatedParamsForCluster:
    dfCols = dfCols + ['tailLength', 'tailLengthFromRecalculatedAngles'] + tailAnglesRecalculated + tailAnglesRecalculated2
  if computeTailAngleParamForCluster or computeMassCenterParamForCluster or computetailAnglesRecalculatedParamsForCluster:
    dfCols = dfCols + ['classification']
  numberOfParameters = len(dfCols)
  
  # Creating an empty dataframe then filling it with the parameters for the whole set of videos
  print("Calculating and storing all parameters:")
  dfParam = pd.DataFrame(columns=dfCols)
  genotypes  = []
  conditions = []
  # This is for the potential reload from previously calculated parameters (stored in the pkl file)
  if keepSpeedDistDurWhenLowNbBends == 1:
    onlyKeepTheseColumns = basicInformation + ['Bout Duration (s)', 'Bout Distance (mm)', 'Bout Speed (mm/s)', 'Angular Velocity (deg/s)', 'IBI (s)']
  else:
    onlyKeepTheseColumns = basicInformation
  removeColumnsWhenAppropriate = [col for col in dfCols if not(col in onlyKeepTheseColumns)]
  # Going through each video listed in the excel file
  for videoId in range(0, len(excelFile)):
    
    if excelFile.loc[videoId, 'path'] == "defaultZZoutputFolder":
      path    = os.path.join(defaultZZoutputFolderPath, excelFile.loc[videoId, 'trial_id'])
    else:
      path    = os.path.join(excelFile.loc[videoId, 'path'], excelFile.loc[videoId, 'trial_id'])
    trial_id  = excelFile.loc[videoId, 'trial_id']
    fq        = excelFile.loc[videoId, 'fq']
    pixelsize = excelFile.loc[videoId, 'pixelsize']
    condition = [val.strip('\'" ') for val in excelFile.loc[videoId, 'condition'][1:-1].split(',')]
    genotype  = [val.strip('\'" ') for val in excelFile.loc[videoId, 'genotype'][1:-1].split(',')]
    include   = [bool(int(val.strip())) for val in excelFile.loc[videoId, 'include'][1:-1].split(',')]
    
    if (not(os.path.exists(os.path.join(path, trial_id + '.pkl'))) or forcePandasDfRecreation):
      if len(supstructOverwrite):
        supstruct = supstructOverwrite
      else:
        with open(os.path.join(path, 'results_' + trial_id + '.txt')) as f:
          supstruct = json.load(f)
      firstFrame = supstruct["firstFrame"]
      lastFrame  = supstruct["lastFrame"]
      with open(os.path.join(path, 'parametersUsedForCalculation.json'), 'w') as outfile:
        print('frameStepForDistanceCalculation', frameStepForDistanceCalculation)
        print('videoFPS', excelFile.loc[videoId, 'fq'])
        print('videoPixelSize', excelFile.loc[videoId, 'pixelsize'])
        json.dump({'frameStepForDistanceCalculation': int(frameStepForDistanceCalculation), 'videoFPS': float(excelFile.loc[videoId, 'fq']), 'videoPixelSize': float(excelFile.loc[videoId, 'pixelsize'])}, outfile)
    else:
      print("reloading previously calculated parameters")
      dfReloadedVid = pd.read_pickle(os.path.join(path, trial_id + '.pkl'))
      if 'BoutDuration' in dfReloadedVid:  # pickle file contains old parameter names, recreate it
        forcePandasDfRecreation = True
        if len(supstructOverwrite):
          supstruct = supstructOverwrite
        else:
          with open(os.path.join(path, 'results_' + trial_id + '.txt')) as f:
            supstruct = json.load(f)
        firstFrame = supstruct["firstFrame"]
        lastFrame  = supstruct["lastFrame"]
        with open(os.path.join(path, 'parametersUsedForCalculation.json'), 'w') as outfile:
          print('frameStepForDistanceCalculation', frameStepForDistanceCalculation)
          print('videoFPS', excelFile.loc[videoId, 'fq'])
          print('videoPixelSize', excelFile.loc[videoId, 'pixelsize'])
          json.dump({'frameStepForDistanceCalculation': int(frameStepForDistanceCalculation), 'videoFPS': float(excelFile.loc[videoId, 'fq']), 'videoPixelSize': float(excelFile.loc[videoId, 'pixelsize'])}, outfile)
      else:
        nbFramesTakenIntoAccountReloaded = max([np.sum(['instaTBF' in param for param in dfReloadedVid.columns.tolist()]), np.sum(['tailAnglesRecalculated' in param for param in dfReloadedVid.columns.tolist()]), np.sum(['instaSpeed' in param for param in dfReloadedVid.columns.tolist()])])
        if nbFramesTakenIntoAccountReloaded < nbFramesTakenIntoAccount:
          raise ValueError("nbFramesTakenIntoAccount was too low when pre-generating the pkl file of a video:" + str(np.sum(['instaTBF' in param for param in dfReloadedVid.columns.tolist()])) + " , " + str(np.sum(['tailAnglesRecalculated' in param for param in dfReloadedVid.columns.tolist()])) + " , " + str(np.sum(['instaSpeed' in param for param in dfReloadedVid.columns.tolist()])) + " ; nbFramesTakenIntoAccount :" + str(nbFramesTakenIntoAccount))
          
        for idx, cond in enumerate(condition):
          indForWellId = (dfReloadedVid['Well_ID'] == idx)
          if include[idx]:
            dfReloadedVid.loc[indForWellId, 'Condition'] = cond
            dfReloadedVid.loc[indForWellId, 'Genotype']  = genotype[idx]
            if minNbBendForBoutDetect > 0:
              ind           = (dfReloadedVid['Number of Oscillations'] < minNbBendForBoutDetect/2)
              dfReloadedVid.loc[ind, removeColumnsWhenAppropriate] = float('NaN')
            if not(genotype[idx] in genotypes):
              genotypes.append(genotype[idx])
            if not(condition[idx] in conditions):
              conditions.append(condition[idx])
          else:
            dfReloadedVid = dfReloadedVid.drop([idx2 for idx2, belongsToWell in enumerate(indForWellId) if belongsToWell])
            if 'level_0' in dfReloadedVid.columns:
              dfReloadedVid = dfReloadedVid.drop(['level_0'], axis=1)
            dfReloadedVid = dfReloadedVid.reset_index()

        dfParam = pd.concat([dfParam, dfReloadedVid])
    
    # Going through each well of the video
    for Well_ID, Cond in enumerate(condition):
      # Not going through this loop if we've already reloaded parameters
      if include[Well_ID] and (not(os.path.exists(os.path.join(path, trial_id + '.pkl'))) or forcePandasDfRecreation):
        print("trial_id:", trial_id, " ; Well_ID:", Well_ID)
        dfParamForWell = pd.DataFrame(columns=dfCols)
        curBoutId = 0
        # Going through each animal present in the well
        for fishId in range(0, len(supstruct["wellPoissMouv"][Well_ID])):
          # Going through each bout performed by the animal
          for NumBout, dataForBout in enumerate(supstruct["wellPoissMouv"][Well_ID][fishId]):
            if not("flag" in dataForBout) or dataForBout["flag"] == 0:
              # Calculating specified parameters for that bout
              if "Bend_Timing" in dataForBout and type(dataForBout["Bend_Timing"]) == list and len(dataForBout["Bend_Timing"]) >= minNbBendForBoutDetect:
                
                # Initial basic information
                
                toPutInDataFrameColumn = basicInformation
                toPutInDataFrame       = [trial_id, Well_ID, NumBout, dataForBout['BoutStart'], dataForBout['BoutEnd'], condition[Well_ID], genotype[Well_ID], (lastFrame - firstFrame) / fq]
                
                if not(genotype[Well_ID] in genotypes):
                  genotypes.append(genotype[Well_ID])
                if not(condition[Well_ID] in conditions):
                  conditions.append(condition[Well_ID])
                
                # Calculates the global kinematic parameters and stores them the dataframe
                
                previousBoutEnd = supstruct["wellPoissMouv"][Well_ID][fishId][NumBout-1]["BoutEnd"] if NumBout > 0 else 0
                listOfGlobalParameters = getGlobalParameters(dataForBout, fq, pixelsize, frameStepForDistanceCalculation, previousBoutEnd, globParam, firstFrame, lastFrame, minimumFrameToFrameDistanceToBeConsideredAsMoving)
                
                toPutInDataFrameColumn = toPutInDataFrameColumn + globParam
                toPutInDataFrame       = toPutInDataFrame       + listOfGlobalParameters
                
                # Raw data
                
                if saveRawDataInAllBoutsSuperStructure:
                  toPutInDataFrameColumn = toPutInDataFrameColumn + rawData
                  toPutInDataFrame       = toPutInDataFrame       + gatherInitialRawData(dataForBout, rawData, fq)
                
                # Tail angles
                
                if getTailAngleSignMultNormalized:
                  toPutInDataFrameColumn = toPutInDataFrameColumn + tailAngles
                  toPutInDataFrame       = toPutInDataFrame       + getTailAngles(dataForBout, smoothingFactor, nbFramesTakenIntoAccount, numberOfBendsIncludedForMaxDetect)
                
                # Calculate "dynamic" tail angle related parameters for clustering
                
                if computeTailAngleParamForCluster:
                
                  toPutInDataFrameColumn = toPutInDataFrameColumn + instaTBF + instaAmp + instaAsym
                  toPutInDataFrame       = toPutInDataFrame       + getDynamicParameters(dataForBout, smoothingFactor, nbFramesTakenIntoAccount, numberOfBendsIncludedForMaxDetect)
                
                # Calculate "dynamic" center of mass related parameters for clustering
                
                if computeMassCenterParamForCluster:
                
                  instaSpeedVal       = getInstaSpeed(dataForBout, nbFramesTakenIntoAccount)
                  instaHeadingDiffVal = getInstaHeadingDiff(dataForBout, nbFramesTakenIntoAccount)
                  instaHorizDisplVal  = getInstaHorizontalDisplacement(dataForBout, nbFramesTakenIntoAccount)
                  
                  toPutInDataFrameColumn = toPutInDataFrameColumn + instaSpeed + instaHeadingDiff + instaHorizDispl
                  toPutInDataFrame       = toPutInDataFrame       + instaSpeedVal + instaHeadingDiffVal + instaHorizDisplVal
                  
                # Recalculates tail angles and calculates 
                
                if computetailAnglesRecalculatedParamsForCluster:

                  tailLength = getTailLength(dataForBout)
                  tailAnglesRecalculatedData  = getTailAngleRecalculated(dataForBout, nbFramesTakenIntoAccount, numberOfBendsIncludedForMaxDetect)
                  tailLengthFromRecalculatedAngles = getTailLength2(tailAnglesRecalculatedData)
                  tailAnglesRecalculatedData2 = getTailAngleRecalculated2(dataForBout, nbFramesTakenIntoAccount, numberOfBendsIncludedForMaxDetect)
                  
                  toPutInDataFrameColumn = toPutInDataFrameColumn + ['tailLength', 'tailLengthFromRecalculatedAngles'] + tailAnglesRecalculated + tailAnglesRecalculated2
                  toPutInDataFrame       = toPutInDataFrame       + [tailLength, tailLengthFromRecalculatedAngles] + tailAnglesRecalculatedData + tailAnglesRecalculatedData2.tolist()
                
                # Adding bout parameters to the dataframe created for the current well
                print("toPutInDataFrameColumn:", len(toPutInDataFrameColumn))
                print("toPutInDataFrame:", len(toPutInDataFrame))
                dfParamForWell.loc[curBoutId, toPutInDataFrameColumn] = toPutInDataFrame
                curBoutId = curBoutId + 1
              
              else:
                
                # Initial basic information
                
                toPutInDataFrameColumn = basicInformation
                toPutInDataFrame       = [trial_id, Well_ID, NumBout, dataForBout['BoutStart'], dataForBout['BoutEnd'], condition[Well_ID], genotype[Well_ID], (lastFrame - firstFrame) / fq]
                
                if not(genotype[Well_ID] in genotypes):
                  genotypes.append(genotype[Well_ID])
                if not(condition[Well_ID] in conditions):
                  conditions.append(condition[Well_ID])
                
                if keepSpeedDistDurWhenLowNbBends:
                  
                  # Calculating the global kinematic parameters and more and stores them the dataframe
                  
                  previousBoutEnd = supstruct["wellPoissMouv"][Well_ID][fishId][NumBout-1]["BoutEnd"] if NumBout > 0 else 0
                  listOfGlobalParameters = getGlobalParameters(dataForBout, fq, pixelsize, frameStepForDistanceCalculation, previousBoutEnd, ['Bout Duration (s)', 'Bout Distance (mm)', 'Bout Speed (mm/s)', 'Angular Velocity (deg/s)', 'IBI (s)'] + addToGlobalParameters, firstFrame, lastFrame, minimumFrameToFrameDistanceToBeConsideredAsMoving)
                  
                  toPutInDataFrameColumn = toPutInDataFrameColumn + ['Bout Duration (s)', 'Bout Distance (mm)', 'Bout Speed (mm/s)', 'Angular Velocity (deg/s)', 'IBI (s)'] + addToGlobalParameters
                  toPutInDataFrame       = toPutInDataFrame       + listOfGlobalParameters
                  
                # Adding bout parameters to the dataframe created for the current well
                
                dfParamForWell.loc[curBoutId, toPutInDataFrameColumn] = toPutInDataFrame
                curBoutId = curBoutId + 1
      
        # Adding dataframe created for the current frame to the dataframe for the whole set of videos
        dfParam = pd.concat([dfParam, dfParamForWell])
  
  # Saving the dataframe  
  if 'level_0' in dfParam.columns:
    dfParam = dfParam.drop(['level_0'], axis=1)
  dfParam = dfParam.reset_index()

  # Gaussian fit outlier removal
  if gaussianFitOutlierRemoval:
    columnsToCheck = [col for col in ('Bout Duration (s)', 'Bout Distance (mm)', 'Number of Oscillations', 'Max absolute TBA (deg.)', 'Absolute Yaw (deg)') if col in dfParam.columns]
    dfParam.loc[(np.abs(scipy.stats.zscore(dfParam[columnsToCheck].astype(float), nan_policy='omit')) > 3).any(axis=1), ~dfParam.columns.isin(basicInformation)] = np.nan

  # Saving dataframe for the whole set of videos as a pickle file
  outfile = open(os.path.join(resFolder, nameOfFile + '.pkl'), 'wb')
  pickle.dump(dfParam,outfile)
  outfile.close()
  
  # Saving dataframe for the whole set of videos as a matlab file
  if saveAllBoutsSuperStructuresInMatlabFormat:
    scipy.io.savemat(os.path.join(resFolder, nameOfFile + '.mat'), {'struct1':dfParam.to_dict("list")})
    
  return [conditions, genotypes, nbFramesTakenIntoAccount, globParam]
